// Intialize the automation sequence
//console.log("Content script loaded for:", window.location.href);

// Ready state handling
let isReady = false;
function setReady() {
  isReady = true;
  chrome.runtime.sendMessage({ action: "contentScriptReady" });
}

if (document.readyState === "complete") {
  setReady();
} else {
  window.addEventListener("load", setReady);
}

// Click elements based on the provided options
async function universalClicker(options) {
  const {
    accountId, // Add accountId to the options
    element,
    textContents,
    attribute,
    selectorText,
    parentElement,
    parentAttribute,
    parentSelectorText,
  } = options;

  // Check if textContents should be replaced with the UserKeys value
  const finalTextContents = await replaceTextContentsIfEmail(
    accountId,
    textContents
  );
  // Simulate human-like click by triggering mouse events
  async function simulateHumanClick(element) {
    const rect = element.getBoundingClientRect();
    const x = rect.left + rect.width / 2;
    const y = rect.top + rect.height / 2;

    element.dispatchEvent(
      new MouseEvent("mousedown", {
        bubbles: true,
        cancelable: true,
        view: window,
        clientX: x,
        clientY: y,
        button: 0,
      })
    );

    element.dispatchEvent(
      new MouseEvent("mouseup", {
        bubbles: true,
        cancelable: true,
        view: window,
        clientX: x,
        clientY: y,
        button: 0,
      })
    );

    element.dispatchEvent(
      new MouseEvent("click", {
        bubbles: true,
        cancelable: true,
        view: window,
        clientX: x,
        clientY: y,
        button: 0,
      })
    );

    //console.log(`Simulated human-like click on element:`, element);
  }

  async function clickElements(elements) {
    for (const el of elements) {
      await simulateHumanClick(el);
    }
  }

  function findElementsWithParent(parentSelector, childSelector) {
    document.querySelectorAll(parentSelector).forEach((parentEl) => {
      const childElements = parentEl.querySelectorAll(childSelector);
      const filteredElements = filterElementsByTextContent(childElements);

      if (filteredElements.length > 0) {
        clickElements(filteredElements);
      } else {
        // console.log("No elements found matching text content within parent.");
      }
    });
  }

  function filterElementsByTextContent(elements) {
    if (finalTextContents) {
      return Array.from(elements).filter((el) =>
        el.textContent
          .trim()
          .toLowerCase()
          .includes(finalTextContents.trim().toLowerCase())
      );
    }
    return elements;
  }

  function buildSelector() {
    if (element && attribute && selectorText) {
      return `${element}[${attribute}*="${selectorText}"]`;
    } else if (element && attribute) {
      return `${element}[${attribute}]`;
    } else if (element) {
      return element;
    }
    return "*"; // Default to selecting all elements if no specific element type is provided
  }

  const childSelector = buildSelector();

  if (parentElement && parentAttribute && parentSelectorText) {
    const parentSelector = `${parentElement}[${parentAttribute}*="${parentSelectorText}"]`;
    findElementsWithParent(parentSelector, childSelector);
  } else {
    const elements = document.querySelectorAll(childSelector);
    const filteredElements = filterElementsByTextContent(elements);

    if (filteredElements.length > 0) {
      await clickElements(filteredElements);
    } else {
      // console.log("No elements found with the specified criteria.");
    }
  }
}
// Function to replace text contents with UserKeys value in case of Google or FB Auth
async function replaceTextContentsIfEmail(accountId, textContents) {
  if (textContents === "UserKeys") {
    // Fetch the UserKeys value from Chrome storage
    console.log(
      `Replacing textContents with UserKeys for account ${accountId}`
    );
    console.log("textContents", textContents);
    const keys = await getStorageData([`UserKeys_${accountId}`]);
    const userKey = keys[`UserKeys_${accountId}`];
    console.log("userKey", userKey);
    if (userKey) {
      return userKey;
    } else {
      console.error(`No UserKeys found for account ${accountId}`);
      return null;
    }
  }

  // Return the original textContents if it's not "email"
  return textContents;
}

// Simulates typing text into elements matching a selector
function simulateTypingElement(element, text) {
  return new Promise((resolve, reject) => {
    if (!element) {
      return reject(new Error("Element not found"));
    }

    element.value = ""; // Clear the input field

    let index = 0;

    const typeCharacter = () => {
      if (index < text.length) {
        const char = text[index];

        // Dispatch keydown event
        element.dispatchEvent(
          new KeyboardEvent("keydown", {
            key: char,
            code: `Key${char.toUpperCase()}`,
            bubbles: true,
          })
        );

        // Update the input field's value
        element.value += char;

        // Dispatch input event
        element.dispatchEvent(new Event("input", { bubbles: true }));

        // Dispatch keyup event
        element.dispatchEvent(
          new KeyboardEvent("keyup", {
            key: char,
            code: `Key${char.toUpperCase()}`,
            bubbles: true,
          })
        );

        index++;
        setTimeout(typeCharacter, Math.random() * 100 + 40); // Mimic human typing speed
      } else {
        resolve(`Typing completed for element: ${element}`);
      }
    };

    typeCharacter();
  });
}

async function simulateTyping(selector, text) {
  const elements = document.querySelectorAll(selector);
  if (elements.length === 0) {
    throw new Error(`Elements not found for selector: ${selector}`);
  }

  for (const element of elements) {
    await simulateTypingElement(element, text);
  }

  return `Typing completed for selector: ${selector}`;
}

// Handles typing user and password credentials into input fields
async function inputCredentials(
  userSelector,
  userText,
  passSelector,
  passText
) {
  try {
    console.log(await simulateTyping(userSelector, userText));
  } catch (error) {
    //console.error(`Error typing user credentials: ${error.message}`);
  }

  try {
    console.log(await simulateTyping(passSelector, passText));
  } catch (error) {
    //console.error(`Error typing password credentials: ${error.message}`);
  }
}

// Universal function to find text content based on various selectors
async function universalFindText(options) {
  const {
    element,
    attribute,
    selectorText,
    textContents,
    slug,
    parentElement,
    parentAttribute,
    parentSelectorText,
  } = options;
  let texts = [];

  function extractTexts(elements) {
    elements.forEach((el) => {
      const text = el.textContent.trim();
      if (
        textContents
          ? text.toLowerCase().includes(textContents.toLowerCase())
          : true
      ) {
        texts.push(text);
      }
    });
  }

  function findElementsWithParent(parentSelector, childSelector) {
    document.querySelectorAll(parentSelector).forEach((parentEl) => {
      extractTexts(parentEl.querySelectorAll(childSelector));
    });
  }

  if (parentElement && parentAttribute && parentSelectorText) {
    findElementsWithParent(
      `${parentElement}[${parentAttribute}*="${parentSelectorText}"]`,
      element || "*"
    );
  } else if (element && attribute && selectorText) {
    extractTexts(
      document.querySelectorAll(`${element}[${attribute}*="${selectorText}"]`)
    );
  } else if (element) {
    extractTexts(document.querySelectorAll(element));
  }

  console.log(`Texts found for ${slug}:`, texts);
  const finalSum = universalFilter(texts, slug);
  compareAndUpdateValue(finalSum, slug);
  return finalSum;
}

// Filters and finds the smallest non-zero number from the text content
function universalFilter(texts, slug) {
  const divideBy100Slugs = ["FC", "Scrooge"];
  let smallestNumber = Infinity;

  texts.forEach((text) => {
    const cleanedText = text.replace(/,/g, "");
    const numberMatch = cleanedText.match(/\d+(\.\d+)?/);

    if (numberMatch) {
      let number = parseFloat(numberMatch[0]);

      if (divideBy100Slugs.includes(slug)) {
        number /= 100;
      }

      if (number !== 0 && number < smallestNumber) {
        smallestNumber = number;
      }
    }
  });

  if (smallestNumber === Infinity) {
    console.log("No valid non-zero numbers found.");
    return null;
  } else {
    smallestNumber = parseFloat(smallestNumber.toPrecision(5));
    console.log(`Smallest non-zero number found: ${smallestNumber}`);
    return smallestNumber;
  }
}

// Function to compare the new value with the stored value and update if necessary
async function compareAndUpdateValue(newValue, accountId) {
  const exemptAccounts = ["Sweep", "Crown"];
  const keys = [
    `value_for_${accountId}`,
    `waitPeriod_for_${accountId}`,
    `timestamp_for_${accountId}`,
    `failCount_for_${accountId}`,
    `manualAlarmSet_${accountId}`,
  ];

  try {
    const result = await getStorageData(keys);
    const currentValue = result[`value_for_${accountId}`];
    const waitPeriodHours = result[`waitPeriod_for_${accountId}`];
    let failCount = result[`failCount_for_${accountId}`] || 0;

    if (waitPeriodHours === undefined) {
      //console.error(`Wait period for '${accountId}' is not defined.`);
      return;
    }

    const isExempt = exemptAccounts.includes(accountId);
    const isValueChanged =
      newValue !== 0 &&
      newValue !== null &&
      (currentValue !== newValue || isExempt);

    if (isValueChanged) {
      // Handle value update and ensure the update completes before moving on
      await updateAccountValue(accountId, newValue);

      // Update the timestamp and manualAlarmSet within storedAccounts
      await updateStoredAccount(accountId, newValue);
    } else {
      // Handle no value change
      handleNoValueChange(result, accountId, waitPeriodHours, failCount);
    }
  } catch (error) {
    // console.error(
    //   `Error retrieving data from storage for '${accountId}': ${error}`
    // );
  }
}

// Function to update account value and timestamp in the storage
async function updateAccountValue(accountId, newValue) {
  const timestamp = new Date().toISOString();
  const newValueObj = {
    [`value_for_${accountId}`]: newValue,
    [`timestamp_for_${accountId}`]: timestamp,
    [`failCount_for_${accountId}`]: 0, // Reset fail count on success
    [`manualAlarmSet_${accountId}`]: false, // Reset the manual alarm flag
  };

  return new Promise((resolve, reject) => {
    chrome.storage.local.set(newValueObj, () => {
      if (chrome.runtime.lastError) {
        reject(chrome.runtime.lastError);
      } else {
        // console.log(`Value for '${accountId}' updated to: ${newValue}`);
        // console.log(`Timestamp for '${accountId}' updated to: ${timestamp}`);
        resolve();
      }
    });
  });
}

// Function to update the storedAccounts object in the storage
async function updateStoredAccount(accountId, newValue) {
  const timestamp = new Date().toISOString();

  return new Promise((resolve, reject) => {
    chrome.storage.local.get("storedAccounts", (result) => {
      if (chrome.runtime.lastError) {
        reject(chrome.runtime.lastError);
        return;
      }

      const storedAccounts = result.storedAccounts || {};
      const accountData = storedAccounts[accountId] || {};
      accountData.timestamp = timestamp;
      accountData.manualAlarmSet = false; // Reset manual alarm flag
      accountData.value = newValue; // Update new value

      storedAccounts[accountId] = accountData;

      chrome.storage.local.set({ storedAccounts }, () => {
        if (chrome.runtime.lastError) {
          reject(chrome.runtime.lastError);
        } else {
          // console.log(
          //   `Stored account for '${accountId}' updated with new timestamp and manualAlarmSet.`
          // );
          resolve();
        }
      });
    });
  });
}

// Function to handle cases where value has not changed
function handleNoValueChange(result, accountId, waitPeriodHours, failCount) {
  const lastTimestamp = new Date(
    result[`timestamp_for_${accountId}`]
  ).getTime();
  const waitPeriodMs = waitPeriodHours * 60 * 60 * 1000;
  const currentTime = Date.now();

  if (currentTime - lastTimestamp >= waitPeriodMs) {
    failCount++;
    // console.log(`Incrementing fail count for '${accountId}' to ${failCount}.`);

    chrome.storage.local.set(
      { [`failCount_for_${accountId}`]: failCount },
      () => {
        // console.log(
        //   `Value for '${accountId}' has not changed. Fail count incremented.`
        // );
      }
    );
  } else {
    // console.log(
    //   `Value for '${accountId}' has not changed, but it's too soon to increment fail count.`
    // );
  }
}

// Handles navigation to a specific URL
function navigateToURL(url) {
  window.location.href = url;
}

// Service worker to handle messages from the content script
chrome.runtime.onMessage.addListener((request, sender, sendResponse) => {
  switch (request.action) {
    case "navigate":
      handleNavigate(request, sendResponse);
      break;

    case "simulateTyping":
      handleSimulateTyping(request, sendResponse);
      break;

    case "clickButton":
      //console.log("Received clickButton action with options:", request);
      handleClickButton(request, sendResponse);
      break;

    case "findVal":
      handleFindVal(request, sendResponse);
      break;
  }
  return true;
});

//Utility functions
function getStorageData(keys) {
  return new Promise((resolve, reject) => {
    chrome.storage.local.get(keys, (result) => {
      if (chrome.runtime.lastError) {
        reject(chrome.runtime.lastError.message);
      } else {
        resolve(result);
      }
    });
  });
}
function handleNavigate(request, sendResponse) {
  console.log("Received navigate action with URL:", request.URL);
  const { URL } = request;
  if (URL) {
    navigateToURL(URL);
    sendResponse({ status: "success", message: `Navigating to ${URL}` });
  } else {
    sendResponse({ status: "error", message: "URL not provided" });
  }
}

function handleSimulateTyping(request, sendResponse) {
  const { userSelector, userText, passSelector, passText } = request;

  inputCredentials(userSelector, userText, passSelector, passText)
    .then(() => sendResponse({ status: "success" }))
    .catch((error) => {
      console.error("Typing simulation failed", error);
      sendResponse({ status: "Failed", error: error.message });
    });
}

function handleClickButton(request, sendResponse) {
  //console.log("request", request);
  //console.log(".data", request.data);

  universalClicker(request)
    .then(() => sendResponse({ status: "success" }))
    .catch((error) => {
      console.error("Error performing click operation:", error);
      sendResponse({ status: "error", message: error.toString() });
    });
}

function handleFindVal(request, sendResponse) {
  console.log("Received findVal action with options:", request);

  universalFindText(request)
    .then((value) => sendResponse({ status: "success", value }))
    .catch((error) => {
      console.error(`FindVal: Error finding value for ${request.slug}`, error);
      sendResponse({ status: "error", message: error.message });
    });
}
function logError(functionName, error) {
  console.error(`Error in ${functionName}:`, error);
  chrome.runtime.sendMessage({
    action: "logError",
    data: { function: functionName, error: error.message },
  });
}
